# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
PQDT, http://pqdtopen.proquest.com.

Config
------

[pqdt]

oai = http://example.endpoint.com/x/y/z

"""

import datetime

import luigi
from gluish.common import Executable
from gluish.format import Gzip
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.oai import pqdt_harvest
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask


class PQDTTask(DefaultTask):
    """PQDT open."""

    TAG = "34"

    def closest(self):
        return monthly(date=self.date)


class PQDTCombine(PQDTTask):
    """Combine files."""

    date = ClosestDateParameter(default=datetime.date.today())
    prefix = luigi.Parameter(default="oai_dc")

    def requires(self):
        return Executable(name="metha-sync", message="https://github.com/miku/metha")

    def run(self):
        url = self.config.get("pqdt", "oai")
        shellout(
            "METHA_DIR={dir} metha-sync -format {prefix} {url}",
            prefix=self.prefix,
            url=url,
            dir=self.config.get("core", "metha-dir"),
        )
        output = shellout(
            "METHA_DIR={dir} metha-cat -format {prefix} {url} | pigz -c > {output}",
            prefix=self.prefix,
            url=url,
            dir=self.config.get("core", "metha-dir"),
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml.gz"))


class PQDTFetch(PQDTTask):
    """
    Rewrite after endpoint behaviour has changed (12/2021).
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        output = pqdt_harvest()
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml"))


class PQDTPrepare(PQDTTask):
    """
    Some postprocessing.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return PQDTFetch()

    def run(self):
        output = shellout(
            """
                 echo '<?xml version="1.0" encoding="UTF-8"?>' | gzip -c >> {output} &&
                 echo '<Records xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">' | gzip -c >> {output} &&
                 xmlcutty -path /OAI-PMH/ListRecords/record < {input} |
                 sed -e 's@<record>@<record xmlns="http://www.openarchives.org/OAI/2.0/">@g' | gzip -c >> {output} &&
                 echo '</Records>' | gzip -c >> {output} && zcat {output} | xmllint --format - | gzip -c | sponge {output}
                 """,
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml.gz"))


class PQDTIntermediateSchema(PQDTTask):
    """
    OAI to intermediate schema.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        # return PQDTCombine(date=self.date)
        return PQDTPrepare(date=self.date)

    def run(self):
        mapdir = "file:///%s" % self.assets("maps/")
        output = shellout(
            """flux.sh {flux} in={input} MAP_DIR={mapdir} | pigz -c > {output}""",
            flux=self.assets("34/flux.flux"),
            mapdir=mapdir,
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj.gz"), format=Gzip)


class PQDTExport(PQDTTask):
    """
    OAI to SOLR schema.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default="solr5vu3")

    def requires(self):
        return {
            "file": PQDTIntermediateSchema(date=self.date),
            "config": AMSLFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout(
            """ span-tag -c {config} {input} > {output} """,
            config=self.input().get("config").path,
            input=self.input().get("file").path,
        )
        output = shellout(
            """ span-export -o {format} {input} > {output} """,
            input=output,
            format=self.format,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj"))



import json

import pymarc

from siskin.conversions import de_listify, imslp_xml_to_marc, osf_to_intermediate


def test_imslp_xml_to_marc():
    example = """<?xml version="1.0"?>
    <document docID="imslpvalsskramstadhans">
      <localClass localClassName="col">imslp</localClass>
      <localClass localClassName="vifa">vifamusik</localClass>
      <identifier identifierEncodingSchema="originalID">valsskramstadhans</identifier>
      <creator>
        <mainForm>Skramstad, Hans</mainForm>
      </creator>
      <title>Vals for pianoforte</title>
      <subject>
        <mainForm>Romantic</mainForm>
      </subject>
      <music_arrangement_of>Piano</music_arrangement_of>
      <url urlEncodingSchema="originalDetailView">http://imslp.org/wiki/Vals_(Skramstad,_Hans)</url>
      <vifatype>Internetressource</vifatype>
      <fetchDate>2018-04-25T00:00:00.01Z</fetchDate>
      <vifaxml><![CDATA[<document docID="imslpvalsskramstadhans"><localClass
          localClassName="col">imslp</localClass><localClass
          localClassName="vifa">vifamusik</localClass><identifier
          identifierEncodingSchema="originalID">valsskramstadhans</identifier><creator><mainForm>Skramstad,
          Hans</mainForm></creator><title>Vals for
          pianoforte</title><subject><mainForm>Romantic</mainForm></subject><music_arrangement_of>Piano</music_arrangement_of><url
          urlEncodingSchema="originalDetailView">http://imslp.org/wiki/Vals_(Skramstad,_Hans)</url><vifatype>Internetressource</vifatype></document>]]></vifaxml>
    </document>
    """

    result = imslp_xml_to_marc(example)

    assert result is not None
    assert isinstance(result, pymarc.Record)

    assert result["001"].value() == "finc-15-dmFsc3NrcmFtc3RhZGhhbnM"
    assert result["100"]["a"] == "Skramstad, Hans"
    assert result["245"]["a"] == "Vals for pianoforte"
    assert result["856"]["u"] == "http://imslp.org/wiki/Vals_(Skramstad,_Hans)"


def test_de_listify():
    cases = (
        (None, None),
        ("", ""),
        ([], None),
        ({1, 2, 3}, 1),
        ([1, 2, 3], 1),
    )
    for v, expected in cases:
        assert de_listify(v) == expected


def test_osf_to_intermediate():
    cases = (
        (None, None),
        ({}, None),
        (
            json.loads(
                """
            {
              "id": "egcsk",
              "type": "preprints",
              "attributes": {
                "date_created": "2021-07-19T07:32:33.252615",
                "date_modified": "2021-07-19T07:42:12.725457",
                "date_published": "2021-07-19T07:41:43.501204",
                "original_publication_date": "2021-02-28T17:00:00",
                "doi": null,
                "title": "Konsep Allah Dalam Teologi Proses",
                "description": "Penulisan karya ilmiah ini dikhususkan untuk membahas mengenai Allah yang dirumuskan dalam teologi proses, yang dicetuskan oleh Alfred Whitehead. Dalam bagian bagian ini penulis menyajikan konsep Allah dalam teologi proses dan bagaimana tanggapan terhadap konsep tersebut secara Alkitabiah Metode penelitian, penulis menggunakan pendekatan metode penelitian kualitatif analisis deskriptif, dengan pendekatan literatur dan tergolong dalam penelitian perpustakaan. Konsep Allah menurut teologi proses adalah Allah yang berproses, tidak berpribadi dan tidak memiliki kedaulatan absolut. Namun pandangan tentang Allah dalam teologi proses adalah suatu kumpulan pengalaman pribadi dan dijadikan sebagai suatu konsep dalam pemikiran manusia. Tanggapan tersebut menunjukan perbandingan dari pola pikir teologi proses mengenai Allah yang menyimpang dan mengarahkan seseorang dalam memahami konsep Allah yang benar sesuai dengan pernyataan Allah m",
                "is_published": true,
                "is_preprint_orphan": false,
                "license_record": {
                  "copyright_holders": [
                    ""
                  ],
                  "year": "2021"
                },
                "tags": [
                  "Gambar",
                  "Respon",
                  "Teologi Proses",
                  "Tuhan"
                ],
                "preprint_doi_created": "2021-07-19T07:42:12.695116",
                "date_withdrawn": null,
                "current_user_permissions": [],
                "public": true,
                "reviews_state": "accepted",
                "date_last_transitioned": "2021-07-19T07:41:43.501204",
                "has_coi": false,
                "conflict_of_interest_statement": null,
                "has_data_links": "no",
                "why_no_data": null,
                "data_links": [],
                "has_prereg_links": "no",
                "why_no_prereg": null,
                "prereg_links": [],
                "prereg_link_info": "",
                "subjects": [
                  [
                    {
                      "id": "584240da54be81056cecaab4",
                      "text": "Arts and Humanities"
                    },
                    {
                      "id": "584240da54be81056cecaa9c",
                      "text": "Religion"
                    },
                    {
                      "id": "584240da54be81056cecaaf5",
                      "text": "Christianity"
                    }
                  ]
                ]
              },
              "relationships": {
                "contributors": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/preprints/egcsk/contributors/",
                      "meta": {}
                    }
                  }
                },
                "bibliographic_contributors": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/preprints/egcsk/bibliographic_contributors/",
                      "meta": {}
                    }
                  }
                },
                "citation": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/preprints/egcsk/citation/",
                      "meta": {}
                    }
                  },
                  "data": {
                    "id": "egcsk",
                    "type": "preprints"
                  }
                },
                "identifiers": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/preprints/egcsk/identifiers/",
                      "meta": {}
                    }
                  }
                },
                "node": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/nodes/uka4p/",
                      "meta": {}
                    },
                    "self": {
                      "href": "https://api.osf.io/v2/preprints/egcsk/relationships/node/",
                      "meta": {}
                    }
                  },
                  "data": {
                    "id": "uka4p",
                    "type": "nodes"
                  }
                },
                "license": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/licenses/563c1cf88c5e4a3877f9e96a/",
                      "meta": {}
                    }
                  },
                  "data": {
                    "id": "563c1cf88c5e4a3877f9e96a",
                    "type": "licenses"
                  }
                },
                "provider": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/providers/preprints/osf/",
                      "meta": {}
                    }
                  },
                  "data": {
                    "id": "osf",
                    "type": "preprint-providers"
                  }
                },
                "files": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/preprints/egcsk/files/",
                      "meta": {}
                    }
                  }
                },
                "primary_file": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/files/60f52a94f1369301d5793a17/",
                      "meta": {}
                    }
                  },
                  "data": {
                    "id": "60f52a94f1369301d5793a17",
                    "type": "files"
                  }
                },
                "review_actions": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/preprints/egcsk/review_actions/",
                      "meta": {}
                    }
                  }
                },
                "requests": {
                  "links": {
                    "related": {
                      "href": "https://api.osf.io/v2/preprints/egcsk/requests/",
                      "meta": {}
                    }
                  }
                }
              },
              "links": {
                "self": "https://api.osf.io/v2/preprints/egcsk/",
                "html": "https://osf.io/egcsk/",
                "preprint_doi": "https://doi.org/10.31219/osf.io/egcsk"
              }
            }"""
            ),
            {
                "abstract": "Penulisan karya ilmiah ini dikhususkan untuk membahas mengenai "
                "Allah yang dirumuskan dalam teologi proses, yang dicetuskan oleh "
                "Alfred Whitehead. Dalam bagian bagian ini penulis menyajikan "
                "konsep Allah dalam teologi proses dan bagaimana tanggapan "
                "terhadap konsep tersebut secara Alkitabiah Metode penelitian, "
                "penulis menggunakan pendekatan metode penelitian kualitatif "
                "analisis deskriptif, dengan pendekatan literatur dan tergolong "
                "dalam penelitian perpustakaan. Konsep Allah menurut teologi "
                "proses adalah Allah yang berproses, tidak berpribadi dan tidak "
                "memiliki kedaulatan absolut. Namun pandangan tentang Allah dalam "
                "teologi proses adalah suatu kumpulan pengalaman pribadi dan "
                "dijadikan sebagai suatu konsep dalam pemikiran manusia. "
                "Tanggapan tersebut menunjukan perbandingan dari pola pikir "
                "teologi proses mengenai Allah yang menyimpang dan mengarahkan "
                "seseorang dalam memahami konsep Allah yang benar sesuai dengan "
                "pernyataan Allah m",
                "authors": [{"rft.aufirst": "Ceria", "rft.aulast": "Ceria"}],
                "doi": "10.31219/osf.io/egcsk",
                "finc.format": "Article",
                "finc.id": "ai-191-egcsk",
                "finc.mega_collection": ["sid-191-col-osf", "Osf"],
                "finc.source_id": "191",
                "languages": ["eng"],
                "rft.atitle": "Konsep Allah Dalam Teologi Proses",
                "rft.date": "2021-07-19",
                "rft.genre": "article",
                "rft.jtitle": "osf",
                "rft.pub": ["OSF Preprints"],
                "subjects": ["Gambar", "Respon", "Teologi Proses", "Tuhan"],
                "url": ["https://doi.org/10.31219/osf.io/egcsk"],
                "x.date": "2021-07-19T07:42:12.695116Z",
            },
        ),
    )
    for v, expected in cases:
        assert osf_to_intermediate(v) == expected



# coding: utf-8
# pylint: disable=C0301,E1101,C0103

# Copyright 2021 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

from __future__ import print_function

import json
import logging
import sys
import tempfile
import time

import requests

import xmltodict

logger = logging.getLogger("siskin")


def pqdt_harvest(
    sleep=0,
    user_agent="Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:47.0) Gecko/20100101 Firefox/47.0",
    endpoint="https://pqdtoai.proquest.com/OAIHandler",
):
    """
    Harvest PQDT and return path to tempfile with raw response XML.

    Background: Both, metha and sickle report bad resumption tokens.

    Oh, what an ugly endpoint; the interface meant to be machine readable
    yields a HTTP 500 when requested with a machine user agent. One may even
    see the conversation that led to this feature.

    Also: user agent header parsing seems case sensitive; recommended
    reading: https://datatracker.ietf.org/doc/html/rfc2616#section-4.2.

    Also: a data endpoint that needs session management, nice.

    TODO: We need to build a single XML file.

    <Records xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    </Records>
    """
    link = "{}?metadataPrefix=oai_dc&verb=ListRecords".format(endpoint)
    cookies = None
    prev_cursor, batch_size = None, 100  # assumed, updated while we observe the cursor

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as tf:
        while True:
            logger.debug(link)
            resp = requests.get(
                link, cookies=cookies, headers={"User-Agent": user_agent}
            )
            if resp.status_code >= 400:
                raise RuntimeError(
                    "harvest failed: {} {}".format(link, resp.status_code)
                )
            logger.debug(
                "retrieved {} {}".format(len(resp.text), resp.text[:50], "...")
            )
            dd = xmltodict.parse(resp.text)
            tf.write(resp.text)
            tf.write("\n")
            try:
                tokenTag = dd["OAI-PMH"]["ListRecords"]["resumptionToken"]
                if tokenTag is None:
                    break
                cursor = int(tokenTag["@cursor"])
                size = int(tokenTag["@completeListSize"])
                token = tokenTag["#text"]
            except KeyError as exc:
                logger.debug(
                    json.dumps(
                        {
                            "text": resp.text,
                            "status": resp.status_code,
                        }
                    )
                )
                raise RuntimeError(
                    "failed to fetch pqdt, unexpected text: {}".format(resp.text)
                )
            else:
                if not token or int(cursor) >= int(size):
                    break
            if resp.cookies:
                cookies = resp.cookies
            if prev_cursor is not None:
                batch_size = cursor - prev_cursor
            if cursor + batch_size >= size:
                logger.info("harvest done")
            else:
                link = r"{}?resumptionToken={}&verb=ListRecords".format(endpoint, token)
                logger.debug("{} {} {}".format(cursor, size, token))
                time.sleep(sleep)
                prev_cursor = cursor

    return tf.name


if __name__ == "__main__":
    formatter = logging.Formatter(
        "[%(asctime)s][%(name)s][%(levelname)-8s] %(message)s"
    )
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel(logging.DEBUG)

    filename = pqdt_harvest()
    print(filename, file=sys.stderr)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301,C0103,W0614,W0401,E0202

# Copyright 2019 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
Helper functions for dealing with OpenURL, refs #5163.
"""

from six.moves.urllib.parse import urlencode


def update_on_value(tdict, tkey, value, first=True):
    """
    Update target dictionary and set tdict[tkey] to value, if value exists and
    is not None. If the value in question is a sequence, first controls,
    whether only the first element should be used as value.
    """
    if value is not None:
        if first and isinstance(value, (list, tuple)):
            if len(value) > 0:
                tdict[tkey] = value[0]
        else:
            tdict[tkey] = value


def openurl_parameters_from_intermediateschema(doc, rfr_id="www.ub.uni-leipzig.de"):
    """
    Given a intermediate schema dictionary, generate OpenURL parameters to be
    used with a (REDI) link resolver.

    Roughly like vufind13/trunk/web/RecordDrivers/AIRecord.php #L1032-1336.
    """
    if not rfr_id:
        raise ValueError("rfr_id must be set")

    if not doc:
        return {}

    params = {
        "url_ver": "Z39.88-2004",
        "ctx_ver": "Z39.88-2004",
        "ctx_enc": "info:ofi/enc:UTF-8",
        "rfr_id": "info:sid/{}:generator".format(rfr_id),
    }
    update_on_value(params, "rft.title", doc.get("rft.atitle"))
    update_on_value(params, "rft.date", doc.get("rft.date"))
    update_on_value(params, "rft.language", doc.get("languages"))

    if doc.get("rft.place") is not None:
        params["rft.place"] = ", ".join(doc.get("rft.place", []))

    genre = doc.get("rft.genre", "article")
    if genre == "proceeding":
        genre = "article"

    if genre == "book":
        params["rft_val_fmt"] = "info:ofi/fmt:kev:mtx:book"
        params["rft.genre"] = "book"
        params["rft.btitle"] = params["rft.title"]
        del params["rft.title"]

        update_on_value(params, "rft_id", doc.get("finc.record_id"))
        update_on_value(params, "rft.btitle", doc.get("rft.btitle"))
        update_on_value(params, "rft.atitle", doc.get("rft.atitle"))
        update_on_value(params, "rft.edition", doc.get("rft.edition"))
        update_on_value(params, "rft.isbn", doc.get("rft.isbn"))
        update_on_value(params, "rft.issn", doc.get("rft.issn"))
        update_on_value(params, "rft.eissn", doc.get("rft.eissn"))
        update_on_value(params, "rft.volume", doc.get("rft.volume"))
        update_on_value(params, "rft.spage", doc.get("rft.spage"))
        update_on_value(params, "rft.epage", doc.get("rft.epage"))
        update_on_value(params, "rft.pages", doc.get("rft.pages"))
        update_on_value(params, "rft.tpages", doc.get("rft.tpages"))
        update_on_value(params, "rft.issue", doc.get("rft.issue"))
        update_on_value(params, "bici", doc.get("bici"))
        update_on_value(params, "rft.series", doc.get("rft.series"))

        authors = doc.get("authors", [])
        if len(authors) > 0:
            author = authors[0]
            update_on_value(params, "rft.au", author.get("rft.au"))
            update_on_value(params, "rft.aucorp", author.get("rft.aucorp"))
            update_on_value(params, "rft.aufirst", author.get("rft.aufirst"))
            update_on_value(params, "rft.auinit", author.get("rft.auinit"))
            update_on_value(params, "rft.auinit1", author.get("rft.auinit1"))
            update_on_value(params, "rft.aulast", author.get("rft.aulast"))
            update_on_value(params, "rft.ausuffix", author.get("rft.ausuffix"))

        update_on_value(params, "rft.genre", doc.get("rft.genre"))
        if doc.get("doi"):
            params["rft_id"] = "info:doi/{}".format(doc.get("doi"))

        update_on_value(params, "rft.pub", doc.get("rft.pub"))

    elif genre == "article":
        del params["rft.title"]

        update_on_value(params, "rft_id", doc.get("finc.record_id"))
        update_on_value(params, "rft.atitle", doc.get("rft.atitle"))
        update_on_value(params, "rft.jtitle", doc.get("rft.jtitle"))
        update_on_value(params, "rft.stitle", doc.get("rft.stitle"))
        update_on_value(params, "rft.date", doc.get("rft.date"))
        update_on_value(params, "rft.issn", doc.get("rft.issn"))
        update_on_value(params, "rft.eissn", doc.get("rft.eissn"))
        update_on_value(params, "rft.ssn", doc.get("rft.ssn"))
        update_on_value(params, "rft.volume", doc.get("rft.volume"))
        update_on_value(params, "rft.spage", doc.get("rft.spage"))
        update_on_value(params, "rft.epage", doc.get("rft.epage"))
        update_on_value(params, "rft.pages", doc.get("rft.pages"))
        update_on_value(params, "rft.issue", doc.get("rft.issue"))
        update_on_value(params, "rft.coden", doc.get("rft.coden"))
        update_on_value(params, "rft.artnum", doc.get("rft.artnum"))
        update_on_value(params, "sici", doc.get("sici"))
        update_on_value(params, "rft.chron", doc.get("rft.chron"))
        update_on_value(params, "rft.quarter", doc.get("rft.quarter"))
        update_on_value(params, "rft.part", doc.get("rft.part"))

        authors = doc.get("authors", [])
        if len(authors) > 0:
            author = authors[0]
            update_on_value(params, "rft.au", author.get("rft.au"))
            update_on_value(params, "rft.aucorp", author.get("rft.aucorp"))
            update_on_value(params, "rft.aufirst", author.get("rft.aufirst"))
            update_on_value(params, "rft.auinit", author.get("rft.auinit"))
            update_on_value(params, "rft.auinit1", author.get("rft.auinit1"))
            update_on_value(params, "rft.aulast", author.get("rft.aulast"))
            update_on_value(params, "rft.ausuffix", author.get("rft.ausuffix"))

        update_on_value(params, "rft.genre", doc.get("rft.genre"))
        if doc.get("doi"):
            params["rft_id"] = "info:doi/{}".format(doc.get("doi"))

    elif genre == "journal":
        update_on_value(params, "rft.issn", doc.get("rft.issn"))
    else:
        params["rft_val_fmt"] = "info:ofi/fmt:kev:mtx:book"
        authors = doc.get("authors", [])
        if len(authors) > 0:
            update_on_value(params, "rft.creator", doc.get("rft.au"))
        update_on_value(params, "rft.pub", doc.get("rft.pub"))
        update_on_value(params, "rft.format", doc.get("finc.format"))
        update_on_value(params, "rft.language", doc.get("languages"))

    return params


def openurl_link_from_intermediateschema(
    doc,
    base="http://www.redi-bw.de/links/ubl?rl_site=ubl&",
    rfr_id="www.ub.uni-leipzig.de",
):
    params = openurl_parameters_from_intermediateschema(doc, rfr_id=rfr_id)
    return "{}{}".format(base, urlencode(params))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301,C0103,W0614,W0401,E0202

# Copyright 2018 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
Helper functions for generating filter config files.

A filterconfig is a JSON file that expresses what an instituition or project
wants (or is allowed) to access from a shared index (expressed as YAML at
[https://git.io/JvC0T](https://git.io/JvC0T)). More examples at
[https://git.io/JvC0L](https://git.io/JvC0L).

"""


class FilterConfig:
    """
    Config accumulator and serializer.
    """

    def __init__(self):
        pass



# coding: utf-8
# pylint: disable=C0103,W0232,C0301,W0703

# Copyright 2018 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
A module for conversions. This module exists to separate orchestration code
(luigi task) from data conversion logic (here), cf. this note from the British
Library:

> The code that performs tasks and the code that orchestrates those tasks were
mixed together in the same source files. This made it very hard to work on
improving any individual task on it’s own, and made testing difficult.

* https://blogs.bl.uk/webarchive/2022/01/ukwa-2021-technical-update.html
"""

import base64
import collections
import json
import logging
import os
import sys
import tarfile
import tempfile
from xml.sax.saxutils import escape, unescape

import pymarc
import requests
import six

import marcx
import xmltodict
from siskin.utils import URLCache

html_escape_table = {'"': "&quot;", "'": "&apos;"}
html_unescape_table = {v: k for k, v in html_escape_table.items()}

marburg_language_mapping = {
    "de": "ger",
    "deu": "ger",
}

logger = logging.getLogger("siskin")


def html_escape(text):
    """
    Escape HTML, see also: https://wiki.python.org/moin/EscapingHtml
    """
    return escape(text, html_escape_table)


def html_unescape(text):
    """
    Unescape HTML, see also: https://wiki.python.org/moin/EscapingHtml
    """
    return unescape(text, html_unescape_table)


def imslp_tarball_to_marc(
    tarball, outputfile=None, legacy_mapping=None, max_failures=30
):
    """
    Convert an IMSLP tarball to MARC binary output file without extracting it.
    If outputfile is not given, write to a temporary location.

    Returns the location of the resulting MARC file.

    A maximum number of failed conversions can be specified with `max_failures`,
    as of 2018-04-25, there were 30 records w/o title.
    """
    if outputfile is None:
        _, outputfile = tempfile.mkstemp(prefix="siskin-")

    stats = collections.Counter()

    with open(outputfile, "wb") as output:
        writer = pymarc.MARCWriter(output)
        with tarfile.open(tarball) as tar:
            for member in tar.getmembers():
                fobj = tar.extractfile(member)
                try:
                    record = imslp_xml_to_marc(
                        fobj.read(), legacy_mapping=legacy_mapping
                    )
                    writer.write(record)
                except ValueError as exc:
                    logger.warn("conversion failed: %s", exc)
                    stats["failed"] += 1
                finally:
                    fobj.close()
                    stats["processed"] += 1

        writer.close()

        if stats["failed"] > max_failures:
            logger.warn(
                "%d records failed, only %d failures allowed",
                stats["failed"],
                max_failures,
            )
            raise RuntimeError("more than %d records failed", max_failures)

        logger.debug(
            "%d/%d records failed/processed", stats["failed"], stats["processed"]
        )

    return outputfile


def imslp_xml_to_marc(s, legacy_mapping=None):
    """
    Convert a string containing a single IMSLP XML record to a pymarc MARC record.

    Optionally take a legacy mapping, associating IMSLP Names with VIAF
    identifiers. Blueprint: https://git.io/vpQPd, with one difference: We
    allow records with no subjects.

    A record w/o title is an error. We check for them, when we add the field.
    """
    dd = xmltodict.parse(s, force_list={"subject", "languages"})

    if legacy_mapping is None:
        legacy_mapping = collections.defaultdict(lambda: collections.defaultdict(str))

    record = marcx.Record(force_utf8=True)
    record.strict = False

    doc = dd["document"]

    record.leader = "     ncs  22        450 "

    identifier = doc["identifier"]["#text"]
    encoded_id = base64.b64encode(six.b(identifier)).rstrip(b"=")
    record.add("001", data="finc-15-{}".format(encoded_id.decode("utf-8")))

    record.add("007", data="cr")

    if doc.get("languages", []):
        langs = [l for l in doc["languages"] if l != "unbekannt"]
        if langs:
            record.add("008", data="130227uu20uuuuuuxx uuup%s  c" % langs[0])
            for l in langs:
                record.add("041", a=l)

    creator = doc["creator"]["mainForm"]
    record.add(
        "100", a=creator, e="cmp", _0=legacy_mapping.get(identifier, {}).get("viaf", "")
    )

    record.add("240", a=legacy_mapping.get(identifier, {}).get("title", ""))

    try:
        record.add("245", a=html_unescape(doc["title"]))
    except KeyError:
        raise ValueError("cannot find title: %s ..." % s[:300])

    record.add("246", a=html_unescape(doc.get("additionalTitle", "")))

    record.add("260", c=doc.get("date", ""))
    record.add("650", y=doc.get("date", ""))
    record.add("500", a=doc.get("abstract", ""))

    for689 = []

    if "subject" in doc:
        if len(doc["subject"]) == 1:
            for689.append(doc["subject"][0]["mainForm"])
        elif len(doc["subject"]) == 2:
            for689.append(doc["subject"][1]["mainForm"])
        else:
            raise ValueError("cannot handle %d subjects", len(doc["subject"]))

        record.add(
            "590", a=for689[0].title(), b=doc.get("music_arrangement_of", "").title()
        )

        for689.append(doc.get("music_arrangement_of", ""))

        for subject in set(for689):
            record.add("689", a=subject.title())

    record.add("700", a=doc.get("contributor", {}).get("mainForm", ""), e="ctb")
    record.add(
        "856", q="text/html", _3="Petrucci Musikbibliothek", u=doc["url"]["#text"]
    )
    record.add("970", c="PN")
    record.add("980", a=identifier, b="15", c="Petrucci Musikbibliothek")
    return record


def olc_to_intermediate_schema(doc):
    """
    Convert a single OLC from SLUB solr into intermediate schema JSON, w/o tags.

    TODO: every mapping should live outside of code for collaborative editing.
    """
    # "interne Bezeichnung":"Fachkatalog"
    internal_to_name = {
        # $ curl "https://is.gd/qPEypK" |
        #     pup 'tr json{}' |
        #     jq -rc '.[] | [.["children"][1]["children"][0]["text"], .["children"][0]["children"][0]["text"]]' |
        #     sed -e 's@","@":"@; s@^\[@@; s@\]@,@' |
        #     grep 'SSG-OLC'
        "SSG-OLC-ALT": "Altertumswissenschaften",
        "SSG-OLC-ANG": "Anglistik",
        "SSG-OLC-ARC": "Architektur",
        "SSG-OLC-ASS": "Afrika südlich der Sahara",
        "SSG-OLC-AST": "Astronomie",
        "SSG-OLC-BAL": "Baltische Länder",
        "SSG-OLC-BEL": "Niederlande",
        "SSG-OLC-BIF": "Bildungsforschung",
        "SSG-OLC-BUB": "Informations-, Buch- und Bibliothekswesen",
        "SSG-OLC-CHE": "Chemie",
        "SSG-OLC-ETH": "Ethnologie",
        "SSG-OLC-FOR": "Forstwissenschaften",
        "SSG-OLC-FRK": "Frankreichkunde und Allgemeine Romanistik",
        # "SSG-OLC-FTH": "Film und Theater", # refs. #16279
        "SSG-OLC-FTH": "Film / Theater",
        "SSG-OLC-GEO": "Geowissenschaften",
        "SSG-OLC-GER": "Germanistik",
        "SSG-OLC-GWK": "Kunst und Kunstwissenschaft",
        "SSG-OLC-HIS": "Geschichte",
        "SSG-OLC-HSW": "Hochschulwesen",
        "SSG-OLC-IBA": "Ibero-Amerika",
        "SSG-OLC-IBL": "Internationale Beziehungen und Länderkunde",
        "SSG-OLC-ITF": "Italienforschung",
        "SSG-OLC-JUR": "Recht",
        "SSG-OLC-KPH": "Klassische Philologie",
        "SSG-OLC-MAT": "Mathematik und Informatik",
        "SSG-OLC-MFO": "Asien und Nordafrika",
        # "SSG-OLC-MKW": "Medien- und Kommunikationswissenschaft", # refs. #16279
        "SSG-OLC-MKW": "Medien- / Kommunikationswissenschaft",
        "SSG-OLC-MUS": "Musikwissenschaft",
        "SSG-OLC-NED": "Niederlande",
        "SSG-OLC-OAS": "Ost- und Südostasien",
        "SSG-OLC-OEB": "Auswahl deutschsprachiger Zeitschriften",
        "SSG-OLC-OEU": "Osteuropa",
        "SSG-OLC-PHA": "Pharmazie",
        "SSG-OLC-PHI": "Philosophie",
        "SSG-OLC-PHY": "Physik",
        "SSG-OLC-POL": "Politikwissenschaft und Friedensforschung",
        "SSG-OLC-PSY": "Psychologie",
        "SSG-OLC-ROK": "Romanischer Kulturkreis",
        "SSG-OLC-SAS": "Südasien",
        "SSG-OLC-SCA": "Nordeuropa",
        "SSG-OLC-SLA": "Slavistik",
        "SSG-OLC-SOW": "Sozialwissenschaften",
        "SSG-OLC-SPO": "Sportwissenschaften",
        "SSG-OLC-SPP": "Spanien und Portugal",
        "SSG-OLC-TEC": "Technik",
        "SSG-OLC-TGE": "Technikgeschichte",
        "SSG-OLC-UMW": "Umwelt",
        "SSG-OLC-VET": "Veterinärmedizin",
        "SSG-OLC-VOR": "Vorderer Orient",
        "SSG-OLC-WIW": "Wirtschaftswissenschaften",
        "SSG-OLC-ZGE": "Zeitgeschichte",
        "SSG-OPC-ANG": "Anglo-American Culture",
        "SSG-OPC-AST": "Astronomie",
        "SSG-OPC-BAL": "Baltische Länder",
        "SSG-OPC-BBI": "Informations-, Buch- und Bibliothekswesen",
        "SSG-OPC-FIN": "Finnougristik",
        "SSG-OPC-FOR": "Forstwissenschaften",
        "SSG-OPC-GEO": "Fachkatalog Geophysik",
        "SSG-OPC-GGO": "Geowissenschaften",
        "SSG-OPC-MAT": "Mathematik",
        "SSG-OPC-PHA": "Pharmazie",
        "SSG-OPC-VOR": "Vorderer Orient",
    }
    olc_format_to_finc_format = {
        "Journal": "Journal",
        "eJournal": "Journal",
        "Article": "Article",
        "electronic Article": "Article",
        "Monograph Series": "Serial",
        "Serial Volume": "Book",
    }
    mega_collections_set = set()
    for internal in doc.get("collection_details", []):
        if not internal.startswith("SSG-"):
            continue
        name = internal_to_name.get(internal)
        if not name:
            continue
        # Philosophie => OLC SSG Philosophie
        mc = "OLC SSG {}".format(name)
        mega_collections_set.add(mc)

    result = {
        "abstract": de_listify(doc.get("abstract")),
        "authors": [{"rft.au": name} for name in doc.get("author2", [])],
        "finc.format": olc_format_to_finc_format.get(
            de_listify(doc.get("format"), "Article")
        ),
        "finc.id": "ai-68-{}".format(doc["id"]),
        "finc.mega_collection": list(mega_collections_set),
        "finc.source_id": "68",
        "languages": doc.get("lang_code", []),
        "rft.atitle": de_listify(doc.get("title", "")),
        "rft.genre": "article",
        "rft.issn": doc.get("issn", []),
        "rft.issue": doc.get("container_issue", ""),
        "rft.jtitle": doc.get("container_title", ""),
        "rft.place": doc.get("rft.place", []),
        "rft.pub": doc.get("publisher", []),
        "rft.volume": doc.get("container_volume", ""),
        "url": doc.get("url", []),
        "x.subtitle": de_listify(doc.get("title_sub", [])),
    }
    date = doc.get("publishDateSort")
    if date and len(date) == 4:
        result["x.date"] = "{}-01-01T00:00:00Z".format(date)
        result["rft.date"] = date
    return result


def de_listify(v, default=None):
    """
    Take any value and returns a string or None, possibly truncating multiple
    values.
    """
    if v is None:
        if default:
            return default
        return v
    if isinstance(v, str):
        if not v and default:
            return default
        return v
    if isinstance(v, set):
        v = list(v)
    if isinstance(v, list):
        if len(v) > 0:
            return v[0]
        else:
            if default:
                return default
            return None
    raise ValueError("cannot de-listify: {}".format(v))


def osf_to_intermediate(osf, force=False, best_effort=True, max_retries=5):
    """
    Convert a document from https://api.osf.io/v2/preprints/?format=json&page=1
    to intermediate schema; see also kiwi:[191], also: 179; refs #20238.
    """
    if not osf:
        return None
    if not "id" in osf:
        raise ValueError("osf record w/o id: {}".format(osf))
    source_id = "179"

    def get_tcid_for_provider(provider):
        # refs 20238#note-8
        provider_tcid_map = {
            "eartharxiv": "sid-{}-col-earth".format(source_id),
            "ecoevorxiv": "sid-{}-col-eco".format(source_id),
            "lawarxiv": "sid-{}-col-law".format(source_id),
            "lissa": "sid-{}-col-lissa".format(source_id),
            "mediarxiv": "sid-{}-col-medi".format(source_id),
            "psyarxiv": "sid-{}-col-psy".format(source_id),
            "socarxiv": "sid-{}-col-soc".format(source_id),
        }
        try:
            return provider_tcid_map[provider]
        except KeyError:
            return "sid-{}-col-{}".format(source_id, provider)

    id = "ai-{}-{}".format(source_id, osf["id"])
    attrs = osf.get("attributes")
    rels = osf.get("relationships")
    if not attrs or not rels:
        raise ValueError("osf record w/o attributes or relationships: {}".format(osf))

    def find_osf_language(doc, with_default="eng"):
        """
        If possible, detect language from abstract and return 3-letter ISO 639
        code. Otherwise return a default.
        """
        if not attrs.get("description"):
            return with_default
        try:
            import cld3
            from iso639 import languages

            result = cld3.get_language(attrs["description"])
            if not result.is_reliable:
                return with_default
            return languages.get(alpha2=result.language).part2b
        except (ImportError, KeyError):
            return with_default

    def fetch_authors(doc, force=False, best_effort=False, max_retries=5):
        """
        Fetch and cache author docs. We will need an extra request, which we'll
        cache locally.

        Example: https://api.osf.io/v2/preprints/egcsk/contributors/
        """
        result = []
        cache = URLCache(
            directory=os.path.join(tempfile.gettempdir(), ".urlcache"),
            max_tries=max_retries,
        )
        url = doc["relationships"]["contributors"]["links"]["related"]["href"]
        try:
            content = cache.get(url, force=force)
        except (RuntimeError, requests.exceptions.ConnectionError) as exc:
            if best_effort:
                logger.debug("[best-effort] skipping: {}".format(exc))
                return result
            raise
        doc = json.loads(content)
        try:
            for item in doc["data"]:
                try:
                    attrs = item["embeds"]["users"]["data"]["attributes"]
                except KeyError as exc:
                    attrs = item["embeds"]["users"]["errors"][0]["meta"]
                result.append(
                    {
                        "rft.aufirst": attrs["given_name"],
                        "rft.aulast": attrs["family_name"],
                    }
                )
        except KeyError as exc:
            logger.debug(
                "failed to find authors for {}: {}".format(
                    exc, json.dumps(doc, indent=4)
                )
            )
        else:
            logger.debug("fetched {}, found {} author(s)".format(url, len(result)))
        return result

    result = {
        "abstract": attrs.get("description", ""),
        "authors": fetch_authors(
            osf, force=force, best_effort=best_effort, max_retries=max_retries
        ),
        "finc.format": "Preprint",
        "finc.id": "ai-{}-{}".format(source_id, osf["id"]),
        "finc.mega_collection": [
            get_tcid_for_provider(rels["provider"]["data"]["id"]),
            rels["provider"]["data"]["id"].capitalize(),
        ],
        "finc.source_id": source_id,
        "languages": [find_osf_language(osf)],
        "doi": osf["links"]["preprint_doi"].replace("https://doi.org/", ""),
        "rft.atitle": attrs["title"],
        "rft.genre": "article",
        "rft.jtitle": rels["provider"]["data"]["id"],
        "rft.pub": ["OSF Preprints"],
        "url": [osf["links"]["preprint_doi"]],
        "subjects": attrs["tags"],
    }

    if "preprint_doi_created" in attrs and attrs["preprint_doi_created"]:
        result.update(
            {
                "x.date": attrs.get("preprint_doi_created", "") + "Z",
                "rft.date": attrs.get("preprint_doi_created", "")[:10],
            }
        )

    return result


def eastview_solr_to_intermediate_schema(
    blob,
    source_id="210",
    tcid="sid-210-col-udbedu",
    collection_name="Universal Database of Social Sciences & Humanities (UDB-EDU)",
):
    """
    Given a string containing raw XML. Each document contains a list of
    documents (100s), refs #12586.
    """
    parsed = xmltodict.parse(blob)
    converted = []
    for doc in parsed["add"]["doc"]:
        dd = dict()
        for f in doc["field"]:
            key, value = f.get("@name"), f.get("#text")
            if not key or not value:
                continue
            dd[key] = value
        # keys: 'id', 'title', 'author', 'source', 'src', 'year', 'number',
        # 'place', 'content', 'volume', 'pages', 'language', 'url', 'medium',
        # 'coll', 'cdate'
        if not dd.get("title") or not dd.get("url"):
            continue
        encoded_id = (
            base64.b64encode(dd["id"].encode("utf-8")).decode("utf-8").rstrip("=")
        )
        authors = [{"rft.au": v} for v in dd.get("author", "").split(",") if v]
        result = {
            "version": "0.9",
            "finc.format": "Article",
            "finc.record_id": dd["id"],
            "finc.id": "ai-{}-{}".format(source_id, encoded_id),
            "finc.source_id": source_id,
            "rft.atitle": dd["title"],
            "rft.jtitle": dd.get("source", ""),
            "url": [dd.get("url", "")],
            "authors": authors,
            "finc.mega_collection": [
                collection_name,
                tcid,
            ],
        }
        if dd.get("content"):
            result["abstract"] = dd["content"][:500] + "..."
        if dd.get("language"):
            result["languages"] = [dd["language"]]
        if dd.get("pages") and dd["pages"] != "-":
            result["rft.pages"] = dd["pages"]
        if dd.get("volume"):
            result["rft.volume"] = dd["volume"]
        if dd.get("number"):
            result["rft.issue"] = dd["number"]
        if dd.get("place"):
            result["rft.place"] = [dd["place"]]
        if dd.get("year"):
            result["x.date"] = "{}-01-01T00:00:00Z".format(dd["year"])
            result["rft.date"] = dd["year"]
        converted.append(result)
    return converted


# Facet fields values from author2_role field across all sources.
# TODO: fill out empty strings, maybe add https://git.io/fNLwY, too.
# TODO: move this out of code into an asset folder or the like.
author_role_mapping = {
    "\xc3\xbcbers": "",
    "\xc3\xbcbersetzer": "",
    "\xc3\xbcbersetzerin": "",
    "arrangeur": "",
    "bearb": "",
    "clb": "",
    "darst": "",
    "donateur": "",
    "ed": "",
    "edtomas": "",
    "fotografin": "",
    "herausgeber": "",
    "herausgeberin": "",
    "hg": "",
    "hrsg": "",
    "illustrator": "",
    "kommentatorin": "",
    "komponist": "",
    "librettist": "",
    "mitarb": "",
    "mitarbeiter": "",
    "mitwirkender": "",
    "photographer": "",
    "pord": "",
    "sonstige": "",
    "t\xc3\xa4nzer": "",
    "textdichter": "",
    "textverf": "",
    "veranstalter": "",
    "verfasser einer einleitung": "",
    "verfasser eines nachworts": "",
    "verfasser eines vorworts": "",
    "verfasser von kommentaren": "",
    "verfasser": "",
    "verfasserin einer einleitung": "",
    "verfasserin eines geleitwortes": "",
    "verfasserin eines nachworts": "",
    "verfasserin eines vorworts": "",
    "verfasserin von erg\xc3\xa4nzendem text": "",
    "verfasserin von zusatztexten": "",
    "verfasserin": "",
    "voc": "",
    "zusammenstellender": "",
    # More, from MAB/151 mostly.
    " [Animation]": "",
    " [Ausstattung]": "",
    " [Darst.]": "",
    " [Drehbuch]": "",
    " [Kamera]": "",
    " [Komp.]": "",
    " [Musik]": "",
    " [Prod.]": "",
    " [Regie]": "",
    " [Schnitt]": "",
    " [Ton]": "",
    ", [Drehbuch]": "",
    "Darst.]": "",
    "Drehbuch]": "",
    "Prod.]": "",
    "[Alt]": "",
    "[Animation]": "",
    "[Assistent]": "",
    "[Assistentin]": "",
    "[Aufnahmeleiter]": "",
    "[Ausstattung]": "",
    "[Autor]": "",
    "[Bariton]": "",
    "[Bass]": "",
    "[Bearb.]": "",
    "[Begr.]": "",
    "[Beleuchtung]": "",
    "[Beratung]": "",
    "[Buch]": "",
    "[Casting]": "",
    "[Co-Produzent]": "",
    "[Computeranimation]": "",
    "[DRehbuch]": "",
    "[Darst.]": "",
    "[Darst.r]": "",
    "[Darst]": "",
    "[Darsteller]": "",
    "[Dehbuch]": "",
    "[Digitale Bildgestaltung]": "",
    "[Dir.]": "",
    "[Donat]": "",
    "[Donateur]": "",
    "[Doz.]": "",
    "[Dozent]": "",
    "[Dozentin]": "",
    "[Dramaturgie]": "",
    "[Drast.]": "",
    "[Drebuch]": "",
    "[Drehbuch]": "",
    "[Flöte]": "",
    "[Fotogr.]": "",
    "[Gitarre]": "",
    "[Herst.-Leitung]": "",
    "[Hrsg.]": "",
    "[Idee]": "",
    "[Ill.]": "",
    "[Inszenierung]": "",
    "[Interpret]": "",
    "[Interviewer]": "",
    "[Kamera]": "",
    "[Klavier]": "",
    "[Komp.]": "",
    "[Kostüm]": "",
    "[Kostüme]": "",
    "[Leitung]": "",
    "[Licht]": "",
    "[Malerei]": "",
    "[Mezzosopran]": "",
    "[Mitarb.]": "",
    "[Mitverf.]": "",
    "[Mitwirkende]": "",
    "[Musik]": "",
    "[Nachr.]": "",
    "[Orgel]": "",
    "[Originalkonzept]": "",
    "[Originalvorl.]": "",
    "[Originalvorlage]": "",
    "[Prod.-Leitung]": "",
    "[Prod.]": "",
    "[Prod]": "",
    "[Produzent]": "",
    "[Projektbetreuer]": "",
    "[Projektbetreuerin]": "",
    "[Recherche]": "",
    "[Red.]": "",
    "[Redaktion SWR]": "",
    "[Redaktion]": "",
    "[Reg.]": "",
    "[Regie]": "",
    "[Saxophon]": "",
    "[Schnitt]": "",
    "[Schöpfer]": "",
    "[Sopran]": "",
    "[Special-Effects]": "",
    "[Spezialeffekte]": "",
    "[Sprecher]": "",
    "[Sprecherin]": "",
    "[Studiengangkoordinator]": "",
    "[Studiengangkoordinatorin]": "",
    "[Studiengangskoordinator]": "",
    "[Studiengangskoordinatorin]": "",
    "[Szenebild]": "",
    "[Szenenbild]": "",
    "[Szenerie]": "",
    "[Szenographie]": "",
    "[Tenor]": "",
    "[Ton]": "",
    "[Tonschnitt]": "",
    "[Violine]": "",
    "[Violoncello]": "",
    "[Visual Effects]": "",
    "[Vorl.]": "",
    "[Vorl]": "",
    "[Vorlage]": "",
    "[Vorr.]": "",
    "[Zeichner]": "",
    "[darst.]": "",
    "[gefeierte Pers.]": "",
    "[guitar]": "",
    "[schnitt]": "",
    "[vocal]": "",
    "[Übers.]": "",
    # Below terms are actually ok, as per:
    # http://www.loc.gov/marc/relators/relaterm.html
    "abr": "abr",
    "acp": "acp",
    "act": "act",
    "adi": "adi",
    "adp": "adp",
    "aft": "aft",
    "anl": "anl",
    "anm": "anm",
    "ann": "ann",
    "ant": "ant",
    "ape": "ape",
    "apl": "apl",
    "app": "app",
    "aqt": "aqt",
    "arc": "arc",
    "ard": "ard",
    "arr": "arr",
    "art": "art",
    "asg": "asg",
    "asn": "asn",
    "ato": "ato",
    "att": "att",
    "auc": "auc",
    "aud": "aud",
    "aui": "aui",
    "aus": "aus",
    "aut": "aut",
    "bdd": "bdd",
    "bjd": "bjd",
    "bkd": "bkd",
    "bkp": "bkp",
    "blw": "blw",
    "bnd": "bnd",
    "bpd": "bpd",
    "brd": "brd",
    "brl": "brl",
    "bsl": "bsl",
    "cas": "cas",
    "ccp": "ccp",
    "chr": "chr",
    "cli": "cli",
    "cll": "cll",
    "clr": "clr",
    "clt": "clt",
    "cmm": "cmm",
    "cmp": "cmp",
    "cmt": "cmt",
    "cnd": "cnd",
    "cng": "cng",
    "cns": "cns",
    "coe": "coe",
    "col": "col",
    "com": "com",
    "con": "con",
    "cor": "cor",
    "cos": "cos",
    "cot": "cot",
    "cou": "cou",
    "cov": "cov",
    "cpc": "cpc",
    "cpe": "cpe",
    "cph": "cph",
    "cpl": "cpl",
    "cpt": "cpt",
    "cre": "cre",
    "crp": "crp",
    "crr": "crr",
    "crt": "crt",
    "csl": "csl",
    "csp": "csp",
    "cst": "cst",
    "ctb": "ctb",
    "cte": "cte",
    "ctg": "ctg",
    "ctr": "ctr",
    "cts": "cts",
    "ctt": "ctt",
    "cur": "cur",
    "cwt": "cwt",
    "dbp": "dbp",
    "dfd": "dfd",
    "dfe": "dfe",
    "dft": "dft",
    "dgg": "dgg",
    "dgs": "dgs",
    "dis": "dis",
    "dln": "dln",
    "dnc": "dnc",
    "dnr": "dnr",
    "dpc": "dpc",
    "dpt": "dpt",
    "drm": "drm",
    "drt": "drt",
    "dsr": "dsr",
    "dst": "dst",
    "dtc": "dtc",
    "dte": "dte",
    "dtm": "dtm",
    "dto": "dto",
    "dub": "dub",
    "edc": "edc",
    "edm": "edm",
    "edt": "edt",
    "egr": "egr",
    "elg": "elg",
    "elt": "elt",
    "eng": "eng",
    "enj": "enj",
    "etr": "etr",
    "evp": "evp",
    "exp": "exp",
    "fac": "fac",
    "fds": "fds",
    "fld": "fld",
    "flm": "flm",
    "fmd": "fmd",
    "fmk": "fmk",
    "fmo": "fmo",
    "fmp": "fmp",
    "fnd": "fnd",
    "fpy": "fpy",
    "frg": "frg",
    "gis": "gis",
    "grt": "grt",
    "his": "his",
    "hnr": "hnr",
    "hst": "hst",
    "ill": "ill",
    "ilu": "ilu",
    "ins": "ins",
    "inv": "inv",
    "isb": "isb",
    "itr": "itr",
    "ive": "ive",
    "ivr": "ivr",
    "jud": "jud",
    "jug": "jug",
    "lbr": "lbr",
    "lbt": "lbt",
    "ldr": "ldr",
    "led": "led",
    "lee": "lee",
    "lel": "lel",
    "len": "len",
    "let": "let",
    "lgd": "lgd",
    "lie": "lie",
    "lil": "lil",
    "lit": "lit",
    "lsa": "lsa",
    "lse": "lse",
    "lso": "lso",
    "ltg": "ltg",
    "lyr": "lyr",
    "mcp": "mcp",
    "mdc": "mdc",
    "med": "med",
    "mfp": "mfp",
    "mfr": "mfr",
    "mod": "mod",
    "mon": "mon",
    "mrb": "mrb",
    "mrk": "mrk",
    "msd": "msd",
    "mte": "mte",
    "mtk": "mtk",
    "mus": "mus",
    "nrt": "nrt",
    "opn": "opn",
    "org": "org",
    "orm": "orm",
    "osp": "osp",
    "oth": "oth",
    "own": "own",
    "pan": "pan",
    "pat": "pat",
    "pbd": "pbd",
    "pbl": "pbl",
    "pdr": "pdr",
    "pfr": "pfr",
    "pht": "pht",
    "plt": "plt",
    "pma": "pma",
    "pmn": "pmn",
    "pop": "pop",
    "ppm": "ppm",
    "ppt": "ppt",
    "pra": "pra",
    "prc": "prc",
    "prd": "prd",
    "pre": "pre",
    "prf": "prf",
    "prg": "prg",
    "prm": "prm",
    "prn": "prn",
    "pro": "pro",
    "prp": "prp",
    "prs": "prs",
    "prt": "prt",
    "prv": "prv",
    "pta": "pta",
    "pte": "pte",
    "ptf": "ptf",
    "pth": "pth",
    "ptt": "ptt",
    "pup": "pup",
    "rbr": "rbr",
    "rcd": "rcd",
    "rce": "rce",
    "rcp": "rcp",
    "rdd": "rdd",
    "red": "red",
    "ren": "ren",
    "res": "res",
    "rev": "rev",
    "rpc": "rpc",
    "rps": "rps",
    "rpt": "rpt",
    "rpy": "rpy",
    "rse": "rse",
    "rsg": "rsg",
    "rsp": "rsp",
    "rsr": "rsr",
    "rst": "rst",
    "rth": "rth",
    "rtm": "rtm",
    "sad": "sad",
    "sce": "sce",
    "scl": "scl",
    "scr": "scr",
    "sds": "sds",
    "sec": "sec",
    "sgd": "sgd",
    "sgn": "sgn",
    "sht": "sht",
    "sll": "sll",
    "sng": "sng",
    "spk": "spk",
    "spn": "spn",
    "spy": "spy",
    "srv": "srv",
    "std": "std",
    "stg": "stg",
    "stl": "stl",
    "stm": "stm",
    "stn": "stn",
    "str": "str",
    "tcd": "tcd",
    "tch": "tch",
    "ths": "ths",
    "tld": "tld",
    "tlp": "tlp",
    "trc": "trc",
    "trl": "trl",
    "tyd": "tyd",
    "tyg": "tyg",
    "uvp": "uvp",
    "vac": "vac",
    "vdg": "vdg",
    "wac": "wac",
    "wal": "wal",
    "wam": "wam",
    "wat": "wat",
    "wdc": "wdc",
    "wde": "wde",
    "win": "win",
    "wit": "wit",
    "wpr": "wpr",
    "wst": "wst",
}



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2016 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
BASE, refs #5994.

* https://www.base-search.net/
* https://api.base-search.net/
* https://en.wikipedia.org/wiki/BASE_(search_engine)
* http://www.dlib.org/dlib/june04/lossau/06lossau.html

Currently (Dec 2018) BASE is provided via SLUB download.

Config
------

[base]

ftp-host = sftp://example.com
ftp-base = /target_data/xyz
ftp-username= admin
ftp-password= s3cr3t
ftp-pattern = *
"""

import datetime
import os

import luigi
from gluish.format import TSV, Gzip
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.common import FTPMirror
from siskin.task import DefaultTask


class BaseTask(DefaultTask):
    """
    Various tasks around base, refs #14947.
    """

    TAG = "126"

    def closest(self):
        return monthly(date=self.date)


class BasePaths(BaseTask):
    """
    Mirror SLUB FTP for base.
    """

    date = luigi.DateParameter(default=datetime.date.today())
    max_retries = luigi.IntParameter(default=10, significant=False)
    timeout = luigi.IntParameter(
        default=20, significant=False, description="timeout in seconds"
    )

    def requires(self):
        return FTPMirror(
            host=self.config.get("base", "ftp-host"),
            base=self.config.get("base", "ftp-base"),
            username=self.config.get("base", "ftp-username"),
            password=self.config.get("base", "ftp-password"),
            pattern=self.config.get("base", "ftp-pattern"),
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class BaseSingleFile(BaseTask):
    """
    Create a single compressed file of tarball.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return BasePaths(date=self.date)

    def run(self):
        with self.input().open() as handle:
            paths = [p.strip().decode("utf-8") for p in handle.readlines()]

        # SOLR has a limit on facet_fields value length.
        sanitize = """
        jq -rc '.author[0] = .author[0][0:4000] |
                .author_sort = .author_sort[0:4000] |
                .author_facet[0] = .author_facet[0][0:4000] |
                .publishDate = (.publishDate // "" | scan("[1-9][0-9][0-9][0-9]")) // ""'
        """

        for path in paths:
            if not path.endswith("finc/latest"):
                continue

            realpath = os.path.realpath(path)
            # Since 12/2019 symlinks might be an absolute path (on the server),
            # which makes this more flaky. Assume, that symlink points to the
            # same directory.
            if realpath.startswith("/"):
                self.logger.debug(
                    "absolute path detected; assuming symlinked file is in the same directory"
                )
                realpath = os.path.join(
                    os.path.dirname(path), os.path.basename(realpath)
                )
            self.logger.debug("found: %s", realpath)
            if not realpath.endswith("gz"):
                raise RuntimeError("want gz (tarball), got: %s", realpath)
            decomp = "unpigz -c"
            if realpath.endswith("tar.gz"):
                decomp = "tar -xOzf"
            self.logger.debug("fixing isil on the fly, refs #20232")
            output = shellout(
                """ {decomp} "{input}" |
                                  {sanitize} |
                                  sed -e 's@"DE-15-FID"@"FID-MEDIEN-DE-15"@' |
                                  doisniffer -S |
                                  pigz -c > {output}""",
                decomp=decomp,
                input=realpath,
                sanitize=sanitize,
            )
            luigi.LocalTarget(output).move(self.output().path)
            break
        else:
            raise RuntimeError("no finc/latest in %s", paths)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ndj.gz"), format=Gzip)



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2019 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
OLC, refs #16279.

A list of (51) collection names and internal identifiers can be found here:
https://is.gd/8DFvOo (GBV).
"""

import datetime
import glob
import json
import os

import luigi
from gluish.format import Zstd
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.conversions import olc_to_intermediate_schema
from siskin.task import DefaultTask
from siskin.utils import sha1obj


class OLCTask(DefaultTask):
    """
    OLC base task.
    """

    TAG = "68"

    # cf. https://is.gd/8DFvOo
    COLLECTIONS = [
        "SSG-OLC-ANG",
        "SSG-OLC-ARC",
        "SSG-OLC-BUB",
        "SSG-OLC-FTH",
        "SSG-OLC-GER",
        "SSG-OLC-GWK",
        "SSG-OLC-KPH",
        "SSG-OLC-MKW",
        "SSG-OLC-MUS",
        "SSG-OLC-PHI",
    ]

    def closest(self):
        return monthly(date=self.date)


class OLCDump(OLCTask):
    """
    Fetch a slice (or all of a SOLR), refs #16279.

    Currently, 9/51 collections are requested, but that might change over time.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        query = " OR ".join(
            ["collection_details:{}".format(c) for c in self.COLLECTIONS]
        )
        output = shellout(
            """
            solrdump -verbose -server {server} -q '{query}' |
            jq -rc . | zstd -c -T0 > {output}
            """,
            query=query,
            server=self.config.get("olc", "solr"),
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        filename = "{}-{}.ndj.zst".format(self.date, sha1obj(self.COLLECTIONS))
        return luigi.LocalTarget(path=self.path(filename=filename), format=Zstd)


class OLCIntermediateSchema(OLCTask):
    """
    Sample convertion.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return OLCDump(date=self.date)

    def run(self):
        with self.input().open() as file:
            with self.output().open("w") as output:
                for i, line in enumerate(file):
                    if i % 100000 == 0:
                        self.logger.debug("@{}".format(i))
                    doc = json.loads(line)
                    result = olc_to_intermediate_schema(doc)
                    output.write(bytes(json.dumps(result), encoding="utf-8"))
                    output.write(b"\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ndj.zst"), format=Zstd)


class OLCIntermediateSchemaDeprecated(OLCTask):
    """
    Run and collect OLC, 68, standalone script. There is licensing information
    added in the standalone script and later during AILicening, too. That
    should be ok.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        """
        Find the most recent output, compress and move into the "siskin" tree.
        """
        shellout(
            "{script} --overwrite --outputformat json",
            script=self.assets(
                "68/68-fincjson",
                ignoremap={1: "most probably ok for skip, due to existing file"},
            ),
        )

        # Find the output file.
        taskdir = os.path.join(self.BASE, self.TAG)
        outputs = sorted(
            glob.glob(os.path.join(taskdir, "68-output-*json")), reverse=True
        )
        if len(outputs) == 0:
            raise RuntimeError(
                "could not find any artifacts for source at {}".format(taskdir)
            )
        path = outputs[0]

        # Compress as AIIntermediateSchema requires all artifacts to be gzip compressed.
        output = shellout("zstd -T0 -c {input} > {output}", input=path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ndj.zst"))



# coding: utf-8
# pylint: disable=C301,C0103,W0201

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
Provides a basic benchmark decorator. Usage:

    @timed
    def run():
        print('Running...')

Just logs the output. Could be extended to send this information to some
service if available.
"""

import functools
import logging
from builtins import object
from timeit import default_timer

logger = logging.getLogger("gluish")


class bcolors(object):
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def red(s):
    return bcolors.FAIL + s + bcolors.ENDC


def yellow(s):
    return bcolors.WARNING + s + bcolors.ENDC


def green(s):
    return bcolors.OKGREEN + s + bcolors.ENDC


class Timer(object):
    """
    A timer as a context manager, slight visual language.
    """

    def __init__(self, green=10, yellow=60):
        """
        Indicate runtimes with colors, green < 10s, yellow < 60s.
        """
        # measures wall clock time, not CPU time!
        # On Unix systems, it corresponds to time.time
        # On Windows systems, it corresponds to time.clock
        self.timer = default_timer

        self.green = green
        self.yellow = yellow

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.end = self.timer()
        self.elapsed_s = self.end - self.start
        self.elapsed_ms = self.elapsed_s * 1000


def timed(method):
    """
    A @timed decorator.
    """

    @functools.wraps(method)
    def _timed(*args, **kwargs):
        """
        Benchmark decorator.
        """
        with Timer() as timer:
            result = method(*args, **kwargs)
        klass = args[0].__class__.__name__
        fun = method.__name__

        msg = "[%s.%s] %0.5f" % (klass, fun, timer.elapsed_s)
        if timer.elapsed_s <= timer.green:
            logger.debug(green(msg))
        elif timer.elapsed_s <= timer.yellow:
            logger.debug(yellow(msg))
        else:
            logger.debug(red(msg))
        return result

    return _timed



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2022 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
Eastview.

* refs. #12586

Config
------

[eastview]

ftp-host = host.name
ftp-username = username
ftp-password = password
ftp-path = /
ftp-pattern = *

"""

import datetime
import json
import os
import tempfile
import zipfile

import luigi
from gluish.common import Executable
from gluish.format import TSV, Gzip, Zstd
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.benchmark import timed
from siskin.common import FTPMirror
from siskin.conversions import eastview_solr_to_intermediate_schema
from siskin.task import DefaultTask


class EastViewTask(DefaultTask):
    TAG = "eastview"

    def closest(self):
        return weekly(date=self.date)


class EastViewPaths(EastViewTask):
    """
    A list of IEEE file paths (via FTP).

    managed-schema
    xmlt.xsl
    xml.zip

    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        host = self.config.get("eastview", "ftp-host")
        username = self.config.get("eastview", "ftp-username")
        password = self.config.get("eastview", "ftp-password")
        base = self.config.get("eastview", "ftp-path")
        pattern = self.config.get("eastview", "ftp-pattern")
        return FTPMirror(
            host=host, username=username, password=password, base=base, pattern=pattern
        )

    @timed
    def run(self):
        output = shellout("sort {input} > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="filelist"), format=TSV)


class EastViewIntermediateSchema(EastViewTask):
    """
    Convert to a single (is) file.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return EastViewPaths(date=self.date)

    def run(self):
        with tempfile.NamedTemporaryFile(delete=False, mode="w") as tf:
            zname = None
            with self.input().open(mode="r") as f:
                for line in f:
                    line = line.decode("utf-8").strip()
                    if not line.endswith("xml.zip"):
                        continue
                    zname = line
                    break
            if zname is None:
                raise ValueError("expected a file named xml.zip")
            with zipfile.ZipFile(zname) as zf:
                for name in zf.namelist():
                    if not name.endswith(".xml"):
                        continue
                    with zf.open(name) as f:
                        docs = eastview_solr_to_intermediate_schema(f.read())
                    self.logger.debug("{} {}".format(name, len(docs)))
                    for doc in docs:
                        tf.write(json.dumps(doc) + "\n")
        output = shellout("pigz -c < {tf} > {output}", tf=tf.name)
        luigi.LocalTarget(output).move(self.output().path)
        os.remove(tf.name)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="json.gz"), format=Gzip)



# coding: utf-8
# pylint: disable=C0301,C0330,W0622,E1101

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
https://www.crossref.org/

CrossRef is an association of scholarly publishers
that develops shared infrastructure to support
more effective scholarly communications.

Our citation-linking network today covers over 68
million journal articles and other content items
(books chapters, data, theses, technical reports)
from thousands of scholarly and professional
publishers around the globe.

Configuration
-------------

[crossref]

doi-blacklist = /tmp/siskin-data/crossref/CrossrefDOIBlacklist/output.tsv

"""

# TODO: see, if
# https://academictorrents.com/details/e4287cb7619999709f6e9db5c359dda17e93d515
# is usable, and how often it gets updated
# https://www.crossref.org/blog/new-public-data-file-120-million-metadata-records/

import datetime
import io
import itertools
import json
import os
import socket
import tempfile
import time
import urllib.error
import urllib.parse
import urllib.request

import requests
from six import string_types

import elasticsearch
import luigi
from gluish.common import Executable
from gluish.format import TSV, Gzip
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import date_range, shellout
from siskin import __version__
from siskin.benchmark import timed
from siskin.mail import send_mail
from siskin.sources.amsl import AMSLFilterConfig, AMSLService
from siskin.task import DefaultTask
from siskin.utils import URLCache, load_set_from_target


class CrossrefTask(DefaultTask):
    """
    Crossref related tasks. See: http://www.crossref.org/
    """

    TAG = "49"

    def closest(self):
        """
        Update frequency.
        """
        return monthly(date=self.date)


class CrossrefRawItems(CrossrefTask):
    """
    Concatenate all harvested items.

    TODO: we can get rid of all tasks before this with span-crossref-sync;

    $ span-crossref-sync -t 30m -mode s -verbose -s 2021-04-27 > /dev/null
    $ time cat $(find ~/.cache/span/crossref-sync/ -type f -name "*gz") >> $(taskoutput CrossrefRawItems)

    or:

    $ span-crossref-sync -t 30m -mode s -verbose -s 2021-04-27 | pigz -c > $(taskoutput CrossrefRawItems)

    Companion cron:

    28 1 * * *  span-crossref-sync -t 30m -mode s -verbose -s 2021-04-27 > /dev/null

    Most updates so far in a single day on 2021-12-22: 10,120,570.

    """

    begin = luigi.DateParameter(
        default=datetime.date(2022, 1, 1),
        description="start of the current crossref update streak",
    )
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        crossref_sync_dir = self.config.get("crossref", "sync-dir")
        # TODO: should this be "-mode s" for sync?
        output = shellout(
            """
            span-crossref-sync -t 30m -s {begin} -c {crossref_sync_dir} | zstd -c -T0 >> {output}
            """,
            begin=self.begin,
            crossref_sync_dir=crossref_sync_dir,
        )  # 22min
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ndj.zst"), format=Zstd)


class CrossrefUniqItems(CrossrefTask):
    """
    Calculate current snapshot via span-crossref-snapshot. About 204m30.910s;
    takes 99m59.783s to extract a value with jq (and parallel).

    With span-crossref-{sync,snapshot} we get 130377934 unique records.

    Cf. https://www.crossref.org/06members/53status.html
    """

    begin = luigi.DateParameter(
        default=datetime.date(2022, 1, 1),
        description="start of the current crossref update streak",
    )
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefRawItems(begin=self.begin, date=self.closest())

    def run(self):
        output = shellout(
            "span-crossref-snapshot -verbose -z -compress-program zstd -o {output} {input}",
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ndj.zst"), format=Zstd)


class CrossrefIntermediateSchema(CrossrefTask):
    """
    Convert to intermediate format via span.
    """

    begin = luigi.DateParameter(
        default=datetime.date(2022, 1, 1),
        description="start of the current crossref update streak",
    )
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "span": Executable(name="span-import", message="http://git.io/vI8NV"),
            "file": CrossrefUniqItems(begin=self.begin, date=self.date),
        }

    @timed
    def run(self):
        output = shellout(
            "span-import -i crossref <(zstd -cd -T0 {input}) | zstd -T0 -c > {output}",
            input=self.input().get("file").path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ndj.zst"), format=Zstd)


class CrossrefCollections(CrossrefTask):
    """
    A collection of crossref collections, refs. #6985. XXX: Save counts as well.
    """

    begin = luigi.DateParameter(
        default=datetime.date(2022, 1, 1),
        description="start of the current crossref update streak",
    )
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "input": CrossrefIntermediateSchema(begin=self.begin, date=self.date),
            "jq": Executable(name="jq", message="https://github.com/stedolan/jq"),
        }

    @timed
    def run(self):
        output = shellout(
            """jq -rc '.["finc.mega_collection"][]?' <(zstd -cd -T0 {input}) | LC_ALL=C sort -S35% -u > {output}""",
            input=self.input().get("input").path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefCollectionsCount(CrossrefTask):
    """
    Report collections and the number of titles per collection.
    """

    begin = luigi.DateParameter(
        default=datetime.date(2022, 1, 1),
        description="start of the current crossref update streak",
    )
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "input": CrossrefIntermediateSchema(begin=self.begin, date=self.date),
            "jq": Executable(name="jq", message="https://github.com/stedolan/jq"),
        }

    @timed
    def run(self):
        output = shellout(
            """jq -rc '.["finc.mega_collection"][]?' <(zstd -cd -T0 {input}) | LC_ALL=C sort -S35% > {output}""",
            input=self.input().get("input").path,
        )

        groups = {}  # Map collection name to its size.
        with open(output) as handle:
            for k, g in itertools.groupby(handle):
                name = k.strip()
                groups[name] = len(list(g))

        with self.output().open("w") as output:
            json.dump(groups, output)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="json"))


class CrossrefCollectionsDifference(CrossrefTask):
    """
    Refs. #7049. Check list of collections against AMSL Crossref collections and
    report difference.

    This task uses an experimental email template in ../assets/mail/7049.tmpl, by
    default no email is sent, only when --to is set to one or more comma
    separated email addresses.
    """

    begin = luigi.DateParameter(
        default=datetime.date(2022, 1, 1),
        description="start of the current crossref update streak",
    )
    date = ClosestDateParameter(default=datetime.date.today())

    to = luigi.Parameter(
        default=None,
        description="email address of recipient, comma separated, if multiple",
    )

    def requires(self):
        return {
            "crossref": CrossrefCollections(begin=self.begin, date=self.date),
            "amsl": AMSLService(date=self.date, name="outboundservices:discovery"),
        }

    @timed
    def run(self):
        if self.to is None:
            self.logger.debug(
                "not sending any email, use --to my@mail.com to send out a report"
            )

        amsl = set()

        with self.input().get("amsl").open() as handle:
            items = json.load(handle)

        for item in items:
            if item["sourceID"] == "49":
                amsl.add(item["megaCollection"].strip())

        self.logger.debug("found %s crossref collections in AMSL" % len(amsl))

        missing_in_amsl = []

        with self.input().get("crossref").open() as handle:
            with self.output().open("w") as output:
                for row in handle.iter_tsv(cols=("name",)):
                    if row.name not in amsl:
                        missing_in_amsl.append(row.name)
                        output.write_tsv(row.name)

        self.logger.debug(
            "%s collections seem to be missing in AMSL", len(missing_in_amsl)
        )

        if self.to is not None:
            # Try to send a message, experimental.
            tolist = [v.strip() for v in self.to.split(",")]

            subject = "%s %s %s" % (
                self.__class__.__name__,
                datetime.datetime.today().strftime("%Y-%m-%d %H:%M"),
                len(missing_in_amsl),
            )

            with io.open(self.assets("mail/7049.tmpl"), encoding="utf-8") as fh:
                template = fh.read()

            message = template.format(
                version=__version__,
                count=len(missing_in_amsl),
                clist="\n".join(missing_in_amsl),
                hostname=socket.gethostname(),
                xref=self.input().get("crossref").path,
                amsl=self.input().get("amsl").path,
            )
            message = message.encode("utf-8")
            send_mail(tolist=tolist, subject=subject, message=message)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefDOIList(CrossrefTask):
    """
    A list of Crossref DOIs.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "input": CrossrefIntermediateSchema(date=self.date),
            "jq": Executable(name="jq", message="https://github.com/stedolan/jq"),
        }

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix="siskin-")
        # process substitution sometimes results in a broken pipe, so extract beforehand
        output = shellout(
            "zstd -cd -T0 {input} > {output}", input=self.input().get("input").path
        )
        shellout(
            """jq -r '.doi?' {input} | grep -o "10.*" 2> /dev/null | LC_ALL=C sort -S50% > {output} """,
            input=output,
            output=stopover,
        )
        os.remove(output)
        output = shellout("""LC_ALL=C sort -S50% -u {input} > {output} """, input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefISSNList(CrossrefTask):
    """
    Just dump a list of all ISSN values. With dups and all.
    """

    begin = luigi.DateParameter(
        default=datetime.date(2022, 1, 1),
        description="start of the current crossref update streak",
    )
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "input": CrossrefUniqItems(begin=self.begin, date=self.date),
            "jq": Executable(name="jq", message="https://github.com/stedolan/jq"),
        }

    @timed
    def run(self):
        output = shellout(
            "jq -r '.ISSN[]?' <(zstd -cd -T0 {input}) 2> /dev/null > {output}",
            input=self.input().get("input").path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefUniqISSNList(CrossrefTask):
    """
    Just dump a list of all ISSN values. Sorted and uniq.
    """

    begin = luigi.DateParameter(
        default=datetime.date(2022, 1, 1),
        description="start of the current crossref update streak",
    )
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefISSNList(begin=self.begin, date=self.date)

    @timed
    def run(self):
        output = shellout("LC_ALL=C sort -u {input} > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefDOIAndISSNList(CrossrefTask):
    """
    A list of Crossref DOIs with their ISSNs.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "input": CrossrefIntermediateSchema(date=self.date),
            "jq": Executable(name="jq", message="https://github.com/stedolan/jq"),
        }

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix="siskin-")
        temp = shellout(
            "zstd -cd -T0 {input} > {output}", input=self.input().get("input").path
        )
        output = shellout(
            """jq -r '[.doi?, .["rft.issn"][]?, .["rft.eissn"][]?] | @csv' {input} | LC_ALL=C sort -S50% > {output} """,
            input=temp,
            output=stopover,
        )
        os.remove(temp)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="csv"))


class CrossrefMembers(CrossrefTask):
    """
    Fetch members information via API.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("span-crossref-members > {output}")
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class CrossrefPrefixList(CrossrefTask):
    """
    DOI prefix and canonical name, refs #13587.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefMembers(date=self.date)

    def run(self):
        """XXX: multiple prefixes per name"""
        output = shellout(
            """jq -rc '.message.items[] |
                          {{"prefix": .prefixes[], "name": .["primary-name"]}} |
                          [.prefix, .name] | @tsv' < {input} > {output}""",
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class CrossrefPrefixMapping(CrossrefTask):
    """
    Map existing collection name to new name.

    Current format: PREFIX CANONICAL-NAME CURRENT-AMSL-MEGACOLLECTION
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "data": CrossrefIntermediateSchema(date=self.date),
            "list": CrossrefPrefixList(date=self.date),
        }

    def run(self):
        namemap = {}

        with self.input().get("list").open() as handle:
            for line in handle:
                line = line.strip()
                if not line:
                    continue
                if "\t" not in line:
                    self.logger.warn("invalid prefix list row: %s", line)
                    continue
                prefix, name = line.split("\t", 1)
                namemap[prefix.strip()] = name.strip()

        self.logger.debug("found %d mappings from prefix to name", len(namemap))

        result = set()

        with self.input().get("data").open() as handle:
            for i, line in enumerate(handle):
                if i % 1000000 == 0:
                    self.logger.debug("[...] at %d", i)
                doc = json.loads(line.decode("utf-8").strip())
                doi = doc.get("doi")
                if not doi:
                    self.logger.warn("document without doi: %s", line)
                    continue
                prefix, _ = doi.split("/", 1)

                # Most records will have a single collection name.
                for mega_collection in doc.get("finc.mega_collection", []):
                    name = namemap.get(prefix)

                    if name is None:
                        # Cache canonical names, if we missed it.
                        resp = requests.get(
                            "https://api.crossref.org/members/%s" % prefix
                        ).json()
                        namemap[prefix] = resp["message"]["primary-name"]
                        name = namemap.get(prefix, "UNDEFINED")
                        self.logger.debug(
                            "namemap now contains %d entries, added %s, %s",
                            len(namemap),
                            prefix,
                            namemap[prefix],
                        )

                    try:
                        name = name.decode("utf-8")
                    except AttributeError as err:
                        pass  # XXX: python 2/3 compat
                    entry = tuple(v for v in (prefix, name, mega_collection))
                    result.add(entry)  # Unique.

        with self.output().open("w") as output:
            self.logger.debug("output at %s", output.name)
            for row in sorted(result):
                output.write_tsv(*row)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefPrefixMappingDiff(CrossrefTask):
    """
    Only emit rows, where canonical and current name differ.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefPrefixMapping(date=self.date)

    def run(self):
        with self.input().open() as handle:
            with self.output().open("w") as output:
                for line in handle:
                    if not isinstance(line, string_types):
                        line = line.decode("utf-8")
                    doi, name, current = line.strip().split("\t")
                    if "{} (CrossRef)".format(name) == current:
                        continue
                    if isinstance(line, string_types):
                        line = line.encode("utf-8")
                    output.write(line)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefDataPrefixList(CrossrefTask):
    """
    List of prefixes actually occuring in the data.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return CrossrefIntermediateSchema(date=self.date)

    def run(self):
        seen = set()

        with self.input().open() as handle:
            for i, line in enumerate(handle):
                if i % 1000000 == 0:
                    self.logger.debug("[...] at %d", i)
                line = line.strip()
                if not line:
                    continue
                doc = json.loads(line)
                doi = doc.get("doi")
                if not doi:
                    self.logger.warn("document without doi: %s", line)
                    continue
                prefix, _ = doi.split("/", 1)
                seen.add(prefix)

        with self.output().open("w") as output:
            for value in sorted(seen):
                output.write_tsv(value)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class CrossrefPrefix13587(CrossrefTask):
    """
    Refs #13587, note #32.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "seen": CrossrefDataPrefixList(date=self.date),
            "mapping": CrossrefPrefixMapping(date=self.date),
        }

    def run(self):
        seen = load_set_from_target(self.input().get("seen"))
        written = set()
        with self.input().get("mapping").open() as handle:
            with self.output().open("w") as output:
                for row in handle.iter_tsv(cols=("prefix", "name", "current")):
                    if row.prefix not in seen:
                        self.logger.debug("not seen: %s", row.prefix)
                        continue
                    cut = row[:2]
                    if tuple(cut) not in written:
                        output.write_tsv(*cut)
                        written.add(tuple(cut))

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
https://www.ceeol.com/, refs #9398.

Currently, there are two XML formats used, a custom CEEOL XML and a slightly
invalid MARCXML. The last update was manually uploaded to a colorful server
@Telekom.

Config:

[ceeol]

disk-dir = /path/to/disk
updates = /path/to/file.xml, /path/to/another.xml, ...
updates-marc = /path/to/file.xml, /path/to/file.xml

"""

from __future__ import print_function

import datetime
import glob
import os
import re
import tempfile

import luigi
from gluish.utils import shellout
from siskin.task import DefaultTask
from gluish.format import Zstd


class CeeolTask(DefaultTask):
    """
    Base task.
    """

    TAG = "53"

    date = luigi.DateParameter(default=datetime.date(2021, 7, 14))


class CeeolIntermediateSchema(CeeolTask):
    """
    Convert XML and concat files into a single intermediate schema output.
    """

    def run(self):
        diskdir = self.config.get("ceeol", "disk-dir")
        files = glob.glob(os.path.join(diskdir, "articles_*xml"))
        if len(files) == 0:
            raise RuntimeError("no file found at {}".format(diskdir))
        _, stopover = tempfile.mkstemp(prefix="siskin-")
        for path in files:
            shellout(
                "span-import -i ceeol {input} | zstd -T0 -c >> {output}",
                input=path,
                output=stopover,
            )
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ndj.zst"), format=Zstd)



# coding: utf-8
# pylint: disable=C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

#
# Adhoc
# =====
#

import datetime
import json
import tempfile

import requests

import luigi
import xlsxwriter
from gluish.format import TSV, Gzip
from gluish.utils import shellout
from siskin.sources.amsl import AMSLCollections, AMSLService
from siskin.sources.crossref import (
    CrossrefCollections,
    CrossrefCollectionsCount,
    CrossrefCollectionsDifference,
)
from siskin.sources.degruyter import DegruyterExport
from siskin.sources.elsevierjournals import ElsevierJournalsExport
from siskin.sources.jstor import JstorExport
from siskin.task import DefaultTask
from siskin.utils import SetEncoder, load_set_from_target


class AdhocTask(DefaultTask):
    """
    Base task for throwaway tasks.
    """

    TAG = "adhoc"


class OADOIDatasetStatusByDOI(AdhocTask):
    """
    Download oadoi.org dataset.
    See release notes at: https://docs.google.com/document/d/1Whfe26oyjTedeW1GGWkq3NADgDbL2R2eXqrJCWS8vcc/edit#

    > These datasets are for non-commercial use. For commercial use, ask us
      about our Support Level Agreement (team@impactstory.org). For more
      information about oaDOI, see http://oadoi.org.
    """

    def run(self):
        url = (
            "https://s3-us-west-2.amazonaws.com/oadoi-datasets/oa_status_by_doi.csv.gz"
        )
        output = shellout("""wget -O "{output}" "{url}" """, url=url)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="csv.gz"), format=Gzip)


class Issue7049(AdhocTask):
    """
    Prepare a few bits about collections, refs #7049.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "amsl": AMSLCollections(date=self.date),
            "crossref": CrossrefCollections(date=self.date),
        }

    def run(self):
        amsl = load_set_from_target(self.input().get("amsl"))
        crossref = load_set_from_target(self.input().get("crossref"))

        with self.output().open("w") as output:
            stats = {
                "amsl": amsl,
                "crossref": crossref,
                "amsl_only": amsl - crossref,
                "crossref_only": crossref - amsl,
                "both": amsl & crossref,
            }
            output.write(json.dumps(stats, cls=SetEncoder))

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="json"))


class Issue7049ExportExcel(AdhocTask):
    """
    Create a simple Excel file, with one sheet per kind.
    """

    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "issue7049": Issue7049(date=self.date),
            "counts": CrossrefCollectionsCount(date=self.date),
        }

    def run(self):
        with self.input().get("counts").open() as handle:
            counts = json.load(handle)

        with self.input().get("issue7049").open() as handle:
            doc = json.load(handle)

        _, stopover = tempfile.mkstemp(prefix="siskin-")
        workbook = xlsxwriter.Workbook(stopover)

        worksheet = workbook.add_worksheet(name="Overview")

        keys = doc.keys()

        worksheet.write(0, 0, "#7049")
        worksheet.write(1, 0, datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

        for i, key in enumerate(keys, start=3):
            worksheet.write(i, 1, key)
            worksheet.write(i, 2, len(doc[key]))

        for _, key in enumerate(keys):
            worksheet = workbook.add_worksheet(name=key)
            for j, item in enumerate(sorted(doc[key])):
                worksheet.write(j, 0, counts.get(item, "NA"))
                worksheet.write(j, 1, item)

        workbook.close()
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xlsx"))


class K10Matches(AdhocTask):
    """
    How many of the ISSN in K10Plus are indexed in finc?

    CSV list: https://is.gd/AW3bCB
    """

    ai = luigi.Parameter(default="http://localhost:8983/solr/biblio")
    finc = luigi.Parameter(default="http://localhost:8983/solr/biblio")

    def run(self):
        r = requests.get("https://is.gd/AW3bCB")
        with self.output().open("w") as output:
            for line in r.text.split("\n"):
                try:
                    issn, count = line.split(",")

                    results = {}

                    results["ai"] = requests.get(
                        "%s/select?q=issn:%s&rows=0&wt=json" % (self.ai, issn)
                    )
                    if results["ai"].status_code != 200:
                        raise RuntimeError("ai reponded with %s" % rr.status_code)

                    results["finc"] = requests.get(
                        "%s/select?q=issn:%s&rows=0&wt=json" % (self.finc, issn)
                    )
                    if results["finc"].status_code != 200:
                        raise RuntimeError("finc reponded with %s" % rr.status_code)

                    output.write_tsv(
                        issn,
                        count,
                        str(results["ai"].json().get("response").get("numFound")),
                        str(results["finc"].json().get("response").get("numFound")),
                    )

                except Exception as exc:
                    self.logger.debug(exc)

    def output(self):
        return luigi.LocalTarget(path=self.path(digest=True), format=TSV)



# coding: utf-8
# pylint: disable=F0401,W0232,E1101,C0103,C0301,W0223,E1123,R0904,E1103

# Copyright 2021 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
Open Citations, https://opencitations.net/.
"""

import tempfile

import luigi
from gluish.format import TSV, Gzip
from gluish.utils import shellout
from siskin.task import DefaultTask


class OCITask(DefaultTask):
    """
    Base task for OCI.
    """

    TAG = "oci"
    download_url = luigi.Parameter(
        default="https://figshare.com/ndownloader/articles/6741422/versions/11"
    )


class OCIDownload(OCITask):
    """
    Download via configured URL, about 30G compressed (2021).
    """

    def run(self):
        output = shellout("""curl -sL "{url}" > {output}""", url=self.download_url)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="zip", digest=True), format=Gzip)


class OCISingleFile(OCITask):
    """
    OCI as a single task.
    """

    def requires(self):
        return OCIDownload()

    def run(self):
        with tempfile.TemporaryDirectory(prefix="siskin-") as tmpdname:
            shellout("unzip -d {dir} {file}", dir=tmpdname, file=self.input().path)
            output = shellout(
                """
                              for f in $(find {dir} -name "*zip"); do
                                  unzip -p $f;
                              done | LC_ALL=C grep -vF 'oci,citing' | zstd -c -T0 > {output}
                              """,
                dir=tmpdname,
            )
            luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(
            path=self.path(ext="ndj.zst", digest=True), format=Gzip
        )


class OCICitingDOI(OCITask):
    """
    List of all citing doi.
    """

    def requires(self):
        return OCISingleFile()

    def run(self):
        output = shellout(
            """zstdcat -T0 "{input}" | cut -d , -f 2 | zstd -c -T0 > {output}
                          """,
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(
            path=self.path(ext="ndj.zst", digest=True), format=Gzip
        )


class OCICitedDOI(OCITask):
    """
    List of all cited doi.
    """

    def requires(self):
        return OCISingleFile()

    def run(self):
        output = shellout(
            """zstdcat -T0 "{input}" | cut -d , -f 3 | zstd -c -T0 > {output}
                          """,
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(
            path=self.path(ext="ndj.zst", digest=True), format=Gzip
        )


class OCICitingDOIUnique(OCITask):
    """
    List of all unique citing doi.
    """

    def requires(self):
        return OCICitingDOI()

    def run(self):
        output = shellout(
            """zstdcat -T0 "{input}" | LC_ALL=C sort -u -S 25% | zstd -c -T0 > {output}
                          """,
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(
            path=self.path(ext="tsv.zst", digest=True), format=Gzip
        )


class OCICitedDOIUnique(OCITask):
    """
    List of all unique cited doi.
    """

    def requires(self):
        return OCICitedDOI()

    def run(self):
        output = shellout(
            """zstdcat -T0 "{input}" | LC_ALL=C sort -u -S 25% | zstd -c -T0 > {output}
                          """,
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(
            path=self.path(ext="tsv.zst", digest=True), format=Gzip
        )


class OCIDOIUnique(OCITask):
    """
    Unique DOI in OCI.
    """

    def requires(self):
        return [OCICitingDOIUnique(), OCICitedDOIUnique()]

    def run(self):
        inputs = " ".join([t.path for t in self.input()])
        output = shellout(
            """zstdcat -T0 {inputs} | LC_ALL=C sort -u -S 25% | zstd -c -T0 > {output}
                              """,
            inputs=inputs,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(
            path=self.path(ext="tsv.zst", digest=True), format=Gzip
        )



# coding: utf-8
# pylint: disable=C0301,E1101,C0103

# Copyright 2021 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

import datetime
import json
import re
import tarfile

import luigi
from gluish.format import TSV, Gzip
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.common import FTPMirror
from siskin.task import DefaultTask


class PerinormTask(DefaultTask):
    """
    Base task for Perinorm, refs #16140.
    """

    TAG = "201"
    current = {
        "date": datetime.date(2021, 4, 15),
        "filename": "basic/perinorm_2021-04-15T11:35:02Z.tar.gz",
    }


class PerinormPaths(PerinormTask):
    """
    Mirror SLUB FTP.
    """

    date = luigi.DateParameter(default=PerinormTask.current["date"])
    max_retries = luigi.IntParameter(default=10, significant=False)
    timeout = luigi.IntParameter(
        default=20, significant=False, description="timeout in seconds"
    )

    def requires(self):
        return FTPMirror(
            host=self.config.get("perinorm", "ftp-host"),
            base=self.config.get("perinorm", "ftp-base"),
            username=self.config.get("perinorm", "ftp-username"),
            password=self.config.get("perinorm", "ftp-password"),
            pattern=self.config.get("perinorm", "ftp-pattern"),
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())


class PerinormExport(PerinormTask):
    """
    Input file is not intermediate schema, so we cannot use span-tag currently;
    tag manually.
    """

    date = luigi.DateParameter(default=PerinormTask.current["date"])

    def requires(self):
        return PerinormPaths(date=self.date)

    def run(self):
        with self.input().open() as file:
            for line in file:
                line = line.strip()
                if not self.current["filename"] in line:
                    continue
                path = line
                break
            else:
                raise RuntimeError(
                    "could not file {} in ftp filelist".format(self.current["filename"])
                )
        self.logger.debug("using {}".format(path))

        # XXX: get these from amsl via: span-amsl-discovery -f -live
        # https://example.amsl.technology | grep "perinorm"
        attachments = {
            "Perinorm (DIN-Normen)": ["DE-Gla1", "DE-Zi4"],
            "Perinorm (DWA-Regelwerk)": ["DE-Gla1"],
            "Perinorm (ISO Standards)": ["DE-Gla1"],
            "Perinorm (VDI-Richtlinien)": ["DE-Gla1", "DE-Zi4"],
        }

        with self.output().open("wb") as output:
            tar = tarfile.open(path, "r:gz")
            for member in tar.getmembers():
                f = tar.extractfile(member)
                for line in f:
                    doc = json.loads(line)
                    for k, isils in attachments.items():
                        if doc["mega_collection"][0] == k:
                            doc["institution"] = isils
                    output.write(json.dumps(doc).encode("utf-8"))
                    output.write(b"\n")

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ndj.gz"), format=Gzip)



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2021 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <czygan@ub.uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
> The place to share your research OSF is a free, open platform to support your
research and enable collaboration.

* https://developer.osf.io/

https://api.osf.io/v2/preprints/?filter[provider]=mediarxiv&format=json&page=1

Refs: #20238.
"""

import datetime
import json
import time

import requests

import luigi
from gluish.format import Zstd
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.conversions import osf_to_intermediate
from siskin.sources.amsl import AMSLFilterConfigFreeze
from siskin.task import DefaultTask


class OSFTask(DefaultTask):
    """
    https://api.osf.io/v2/preprints/?filter[provider]=mediarxiv&format=json&page=1
    """

    TAG = "191"

    def closest(self):
        return weekly(date=self.date)


class OSFDownload(OSFTask):
    """
    Download OSF metadata via API. API is kind of slow and flaky.

    Pagination docs: https://jsonapi.org/format/#fetching-pagination

    Retrieval of full set takes: 117m13.266s - about 325M, 86651 docs.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        page = 1
        max_retries = 20  # a "global" retry budget
        sleep_s = 60
        with self.output().open("w") as output:
            while True:
                link = "https://api.osf.io/v2/preprints/?page={}&page[size]=100".format(
                    page
                )
                self.logger.debug("osf: {}".format(link))
                resp = requests.get(link)
                if resp.status_code == 404:
                    break
                if resp.status_code != 200:
                    if max_retries > 0:
                        time.sleep(sleep_s)
                        max_retries -= 1
                        continue
                    else:
                        raise RuntimeError(
                            "osf api failed with {}".format(resp.status_code)
                        )
                self.logger.debug(
                    "fetched {} from {}: {}".format(
                        len(resp.text), link, resp.text[:40]
                    )
                )
                output.write(resp.text)
                output.write("\n")
                page += 1

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="json"))


class OSFIntermediateSchema(OSFTask):
    """
    Convert to intermediate schema (stub).
    """

    date = ClosestDateParameter(default=datetime.date.today())
    max_retries = luigi.IntParameter(
        default=5, description="number of HTTP request retries", significant=False
    )
    encoding = luigi.Parameter(default="utf-8", significant=False)

    def requires(self):
        return OSFDownload()

    def run(self):
        i = 0
        bNL = "\n".encode(self.encoding)
        with self.output().open("w") as output:
            with self.input().open() as f:
                for line in f:
                    resp = json.loads(line)
                    for doc in resp["data"]:
                        result = osf_to_intermediate(doc, max_retries=self.max_retries)
                        if i % 1000 == 0:
                            self.logger.debug("converted {} docs".format(i))
                        output.write(json.dumps(result).encode(self.encoding))
                        output.write(bNL)
                        i += 1

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="json.zst"), format=Zstd)


class OSFExport(OSFTask):
    """
    OSF should be a daily/weekly updated source, so we generate a solr
    importable file directly.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {
            "data": OSFIntermediateSchema(),
            "config": AMSLFilterConfigFreeze(),
        }

    def run(self):
        output = shellout(
            """
            unpigz -c {input} |
            span-tag -unfreeze {config} |
            span-export |
            zstd -c -T0 > {output}""",
            config=self.input().get("config").path,
            input=self.input().get("data").path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="json.zst"))



from siskin.openurl import openurl_parameters_from_intermediateschema


def test_openurl_from_intermediateschema():
    cases = (
        ("empty doc", {}, {}),
        (
            "title only",
            {"rft.atitle": "empty doc"},
            {
                "ctx_enc": "info:ofi/enc:UTF-8",
                "ctx_ver": "Z39.88-2004",
                "rfr_id": "info:sid/www.ub.uni-leipzig.de:generator",
                "rft.atitle": "empty doc",
                "url_ver": "Z39.88-2004",
            },
        ),
        (
            "title and date",
            {
                "rft.atitle": "title and date",
                "rft.date": "2018-10-10",
            },
            {
                "ctx_enc": "info:ofi/enc:UTF-8",
                "ctx_ver": "Z39.88-2004",
                "rfr_id": "info:sid/www.ub.uni-leipzig.de:generator",
                "rft.date": "2018-10-10",
                "rft.atitle": "title and date",
                "url_ver": "Z39.88-2004",
            },
        ),
        (
            "title and date, language",
            {
                "languages": ["eng", "fra"],
                "rft.atitle": "title and date, language",
                "rft.date": "2018-10-10",
            },
            {
                "ctx_enc": "info:ofi/enc:UTF-8",
                "ctx_ver": "Z39.88-2004",
                "rfr_id": "info:sid/www.ub.uni-leipzig.de:generator",
                "rft.date": "2018-10-10",
                "rft.language": "eng",
                "rft.atitle": "title and date, language",
                "url_ver": "Z39.88-2004",
            },
        ),
        (
            "title and date, language, book",
            {
                "languages": ["eng", "fra"],
                "rft.atitle": "Hello",
                "rft.date": "2018-10-10",
                "rft.genre": "book",
            },
            {
                "ctx_enc": "info:ofi/enc:UTF-8",
                "ctx_ver": "Z39.88-2004",
                "rfr_id": "info:sid/www.ub.uni-leipzig.de:generator",
                "rft.atitle": "Hello",
                "rft.btitle": "Hello",
                "rft.date": "2018-10-10",
                "rft.genre": "book",
                "rft.language": "eng",
                "rft_val_fmt": "info:ofi/fmt:kev:mtx:book",
                "url_ver": "Z39.88-2004",
            },
        ),
        (
            "crossref-1",
            {
                "finc.format": "ElectronicArticle",
                "finc.mega_collection": [
                    "Springer Science + Business Media (CrossRef)"
                ],
                "finc.id": "ai-49-aHR0cDovL2R4LmRvaS5vcmcvMTAuMTAxNi9qLm51cnguMjAwNi4wNS4wMjU",
                "finc.source_id": "49",
                "ris.type": "EJOUR",
                "rft.atitle": "An Analysis of Correlations Among 4 Outcome Scales Employed in Clinical Trials of Patients With Major Depressive Disorder",
                "rft.epage": "412",
                "rft.genre": "article",
                "rft.issn": ["1545-5343"],
                "rft.issue": "3",
                "rft.jtitle": "NeuroRX",
                "rft.tpages": "2",
                "rft.pages": "411-412",
                "rft.pub": ["Springer Science + Business Media"],
                "rft.date": "2006-07-01",
                "x.date": "2006-07-01T00:00:00Z",
                "rft.spage": "411",
                "rft.volume": "3",
                "authors": [
                    {"rft.aulast": "JIANG", "rft.aufirst": "Q"},
                    {"rft.aulast": "AHMED", "rft.aufirst": "S"},
                    {"rft.aulast": "PEDERSEN", "rft.aufirst": "R"},
                    {"rft.aulast": "MUSGNUNG", "rft.aufirst": "J"},
                    {"rft.aulast": "ENTSUAH", "rft.aufirst": "R"},
                ],
                "doi": "10.1016/j.nurx.2006.05.025",
                "languages": ["eng"],
                "url": ["http://dx.doi.org/10.1016/j.nurx.2006.05.025"],
                "version": "0.9",
                "x.subjects": ["Pharmacology (medical)"],
                "x.type": "journal-article",
            },
            {
                "ctx_enc": "info:ofi/enc:UTF-8",
                "ctx_ver": "Z39.88-2004",
                "rfr_id": "info:sid/www.ub.uni-leipzig.de:generator",
                "rft.atitle": "An Analysis of Correlations Among 4 Outcome Scales Employed in Clinical Trials of Patients With Major Depressive Disorder",
                "rft.aufirst": "Q",
                "rft.aulast": "JIANG",
                "rft.date": "2006-07-01",
                "rft.epage": "412",
                "rft.genre": "article",
                "rft.issn": "1545-5343",
                "rft.issue": "3",
                "rft.jtitle": "NeuroRX",
                "rft.language": "eng",
                "rft.pages": "411-412",
                "rft.spage": "411",
                "rft.volume": "3",
                "rft_id": "info:doi/10.1016/j.nurx.2006.05.025",
                "url_ver": "Z39.88-2004",
            },
        ),
    )

    for _, doc, want in cases:
        result = openurl_parameters_from_intermediateschema(doc)
        assert result == want



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
"""
datacite.org

> Locate, identify, and cite research data with the leading global provider of DOIs for research data.

Config:

[core]

metha-dir = /path/to/dir

"""

import datetime

import luigi
from gluish.common import Executable
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout
from siskin.task import DefaultTask


class DataciteTask(DefaultTask):
    """
    Base task.
    """

    def closest(self):
        return monthly(date=self.date)


class DataciteCombine(DataciteTask):
    """
    Single file dump.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default="http://oai.datacite.org/oai", significant=False)
    prefix = luigi.Parameter(default="oai_dc", significant=False)

    def requires(self):
        return Executable(name="metha-sync", message="https://github.com/miku/metha")

    def run(self):
        shellout(
            "METHA_DIR={dir} metha-sync -format {prefix} {url}",
            prefix=self.prefix,
            url=self.url,
            dir=self.config.get("core", "metha-dir"),
        )
        output = shellout(
            "METHA_DIR={dir} metha-cat -root Records -format {prefix} {url} | pigz -c > {output}",
            prefix=self.prefix,
            url=self.url,
            dir=self.config.get("core", "metha-dir"),
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml.gz"))


class DataciteIntermediateSchema(DataciteTask):
    """
    Experimental.
    """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return DataciteCombine(date=self.date)

    def run(self):
        mapdir = "file:///%s" % self.assets("maps/")
        output = shellout(
            """flux.sh {flux} in={input} MAP_DIR={mapdir} | pigz -c > {output}""",
            flux=self.assets("datacite/flux.flux"),
            mapdir=mapdir,
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj.gz"))


class DataciteExport(DataciteTask):
    """
    Export to various formats
    """

    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default="solr5vu3", description="export format")

    def requires(self):
        return DataciteIntermediateSchema(date=self.date)

    def run(self):
        output = shellout(
            "span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}",
            format=self.format,
            input=self.input().path,
        )
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            "solr5vu3": "ldj.gz",
            "formeta": "form.gz",
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, "gz")))
